﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;

namespace HeuristicLab.Analysis.AlgorithmBehavior {
  public static class AlgorithmBehaviorHelpers {
    public static bool IsSubtour(IntArray tour, Permutation permutation) {
      // determine starting position for subtour match
      int idx = permutation.Select((x, index) => new { Value = x, Index = index }).Single(x => x.Value == tour[0]).Index;
      bool isSubtour = true;
      for (int i = 1; i < tour.Length; i++) {
        if (tour[i] == permutation[(idx + 1) % permutation.Length]) { // check right side
          idx = (idx + 1) % permutation.Length;
        } else if (tour[i] == permutation[(idx - 1 + permutation.Length) % permutation.Length]) { // check left side
          idx = (idx - 1 + permutation.Length) % permutation.Length;
        } else {
          isSubtour = false;
          break;
        }
      }
      return isSubtour;
    }

    public static List<IntArray> ExtractSubtours(Permutation permutation) {
      var subtours = new List<IntArray>();
      for (int i = 2; i <= permutation.Count() / 2; i++) { // increase schema length from 2 to n/2
        for (int j = 0; j < permutation.Count(); j++) { // visit all positions in the permutation
          var schema = new IntArray(i);
          for (int k = 0; k < i; k++) schema[k] = permutation[(j + k) % permutation.Length]; // copy edge to schema
          subtours.Add(schema);
        }
      }
      return subtours;
    }

    public static string CalculateBrokenInheritanceOccurrences(IntArray schema, GenealogyGraphNode individual) {
      ISet<int> generations = new SortedSet<int>();
      if (IsMatch(schema, ((Permutation)individual.Data).ToArray())) generations.Add(0);
      var filteredDescendants = individual.Descendants().Where(x => IsMatch(schema, ((Permutation)x.Data).ToArray()));
      if (filteredDescendants.Any()) {
        foreach (var descendant in filteredDescendants) {
          generations.Add((int)descendant.Rank);
        }
      }

      StringBuilder sb = new StringBuilder();
      if (generations.Any()) {
        for (int i = 0; i < generations.Count; i++) {
          if (i != 0) sb.Append(", ");
          sb.Append(generations.ElementAt(i));
        }
      } else {
        sb.Append("none");
      }
      return sb.ToString();
    }

    // attention: brute force matching
    // should be replaced with KMP, BM or RK
    public static bool IsMatch(IntArray pattern, int[] tour) {
      bool match = false;
      for (int i = 0; i <= tour.Length - pattern.Length && !match; i++) {
        for (int j = 0; j < pattern.Length && (match = pattern[j] == tour[i + j] || pattern[j] == -1); j++) ;
      }
      return match;
    }
  }
}
